/* -*- mode: c++; c-basic-offset: 4; -*- */
#include "segment.hh"
#include "buffer.hh"
#include "test_endian.hh"
#include "gds_atomic.h"
#include <stdexcept>
#include <iostream>
#include <cstring>

using namespace sends;
using namespace std;

static test_endian bOrder;

//======================================  Segment default constructor
segment::segment(void) 
    : mBufClass(-1), mData(0)
{
    clear();
}

//======================================  Segment destructor
segment::~segment(void) {
    release();
}

//======================================  Reset the segment maximum time
void
segment::cap(gps_type tMax) {
    if (tMax > mTMax || tMax < mStop || tMax < mReserved) {
	cerr << "Segment cap error, tMax: " << tMax << endl;
	dump_header(cerr);
	dump(cerr);
	throw invalid_argument("Invalid cap time");
    } else {
	cout << "Segment for chan: " << mChannel << " times " << mStart 
	     << ":" << mTMax << " capped at: " << tMax << endl;
    }
    mTMax = tMax;
}

//======================================  Clear the segment (retain the buffer)
void
segment::clear(void) {
    mChannel  =  ~0;
    mStart    =   0;
    mStop     =   0;
    mReserved =   0;
    mTMax     =   0;
    mRate     = 0.0;
    mWordType = data_type::kByte1;
    mUseCount =   0;
    mBias     = 0.0;
    mScale    = 1.0;
    mUnits.clear();
}

//======================================  Calculate data position
index_type
segment::dataIndex(gps_type gps) const {
    double rw = mRate;
    int wsize = mWordType.size();
    if (gps < mStart || rw <= 0 || wsize <= 0) {
	cerr << "Start: " << mStart << " t: "    << gps 
	     << " rate: " << mRate  << " type: " << mWordType.str() << endl;
	throw runtime_error("Segment info not valid");
    }
    return static_cast<index_type>(rw * (gps - mStart) + 0.5) * wsize;
}

//======================================  Dump a segment
std::ostream&
segment::dump(std::ostream& out) const {
    out << mChannel  << "  " << mBufClass << "  " << mStart << "  " 
	<< mStop     << "  " << mReserved << "  " << mTMax  << "  " 
	<< mRate     << "  " << mWordType.str() << "  " << mUseCount << endl;
    return out;
}

//======================================  Print a segment dump header
std::ostream&
segment::dump_header(std::ostream& out) {
    out << "chan  bClass  start  stop reserved max rate size use" << endl;
    return out;
}

//======================================  Extend the segment data field.
bool
segment::extend(const void* data, count_type inx, count_type length) {
    char* p = refData(mStop);
    count_type wsize = mWordType.size();
    count_type lBytes = length * wsize;
    mData->push(lBytes);
    const char* in = reinterpret_cast<const char*>(data) + inx * wsize;

    //-----------------------------------  memcpy if big-endian or byte data
    if (bOrder.big_end() || wsize < 2) {
	memcpy(p, in, lBytes);

    //-----------------------------------  Swap bytes.
    } else {
	for (count_type i=0; i<length; ++i) {
	    for (count_type j=wsize; j--;) {
		*p++ = in[j];
	    }
	    in += wsize;
	}
    }
    int dt = int(length / mRate + 0.5);
    return fill_increment(dt);
}

//======================================  Lock/Unlock the fill mutex
bool
segment::fill_increment(int dt) {
    fill_test("fill_increment");
    bool result = false;
    gps_type sv_reserve = mReserved;
#ifdef USE_BUILTIN_ATOMICS
    if (ATOMIC_PREADD(mStop, dt) == sv_reserve) {
	mFillMutex.unlock();
	result = true;
    }
#else
    mUseMutex.lock();
    mStop += dt;
    result = (mStop == sv_reserve);
    if (result) mFillMutex.unlock();
    mUseMutex.unlock();
#endif
    return result;
}

//======================================  Lock/Unlock the fill mutex
void
segment::fill_lock(void) {
    mFillMutex.writelock();
}

//======================================  Fill failed... unreserve the data
void
segment::fill_failed(void) {
    fill_test("fill_failed");
    mReserved = mStop;
    mFillMutex.unlock();
}

void
segment::fill_test(const string& name) {
    if (mFillMutex.trylock()) {
	mFillMutex.unlock();
	throw logic_error(name + " called with mutex unlocked");
    }
}

void
segment::fill_wait(void) {
    mFillMutex.lock();
    mFillMutex.unlock();
}

//======================================  Get a pointer to the data
const char* 
segment::refData(gps_type gps) const {
    return mData->data() + dataIndex(gps);
}

char* 
segment::refData(gps_type gps) {
    return mData->data() + dataIndex(gps);
}

//======================================  Clear the segment
void
segment::release(void) {
    if (mData) {
	mData->free();
	mData =   0;
    }
    mBufClass =  -1;
    clear();
}

//======================================  Reserve space in the data field.
bool
segment::reserve(gps_type start, gps_type stop) {
    //----------------------------------  Check that start and stop are valid.
    if (stop > mTMax || start >= stop) return false;
    if (start != mStop) return false;

#ifdef USE_BUILTIN_ATOMICS
    if (!ATOMIC_CMP_SWAP(mReserved, start, stop)) return false;
#else
    {
	thread::semlock lock(mUseMutex);
	if (mReserved != start) return false;
	mReserved = stop;
    }
#endif 
    fill_lock(); // Beware of hangup... Other locks?
    return true;
}

//======================================  Get a buffer
void 
segment::set_buffer(buffer* bufpt, int bufclass) {
    release();
    mBufClass = bufclass;
    mData     = bufpt;
}

//======================================  Initialize the segment
void 
segment::set_segment(chan_index chan, double rate, 
		     data_type typ, gps_type start) {
    if (!mData) throw runtime_error("Segment buffer not allocated");
    mChannel  = chan;
    mRate     = rate;
    mWordType = typ;
    mStart    = start;
    mStop     = start;
    mReserved = start;
    long nSec = long((mData->capacity()/typ.size())/mRate);
    if (mRate < 1.0) {
	int step = int(1.0/mRate + 0.5);
	nSec -= nSec%step;
    }
    mTMax = mStart + nSec;
    set_units(0.0, 1.0, "");
}

//======================================  Initialize the segment
void 
segment::set_units(double bias, double scale, const std::string& units) {
    mBias  = bias;
    mScale = scale;
    mUnits = units;
}

//======================================  Increment the use count
segment::count_type
segment::unuse(void) {
    if (!mUseCount) throw logic_error("Decrement use count of unused segment");
#ifdef USE_BUILTIN_ATOMICS
    return ATOMIC_PREDEC(mUseCount);
#else
    thread::semlock lock(mUseMutex);
    return --mUseCount;
#endif
}

//======================================  Increment the use count
segment::count_type
segment::use(void) {
#ifdef USE_BUILTIN_ATOMICS
    return ATOMIC_POSTINC(mUseCount);
#else
    thread::semlock lock(mUseMutex);
    return mUseCount++;
#endif
}
